@@ -3,6 +3,8 @@ package ttlcache
 import (
 	"sync"
 	"time"
+
+	"golang.org/x/sync/singleflight"
 )
 
 // CheckExpireCallback is used as a callback for an external check on item expiration
@@ -18,12 +20,24 @@ type ExpireReasonCallback func(key string, reason EvictionReason, value interfac
 // LoaderFunction can be supplied to retrieve an item where a cache miss occurs. Supply an item specific ttl or Duration.Zero
 type LoaderFunction func(key string) (data interface{}, ttl time.Duration, err error)
 
+// SimpleCache interface enables a quick-start. Interface for basic usage.
+type SimpleCache interface {
+	Get(key string) (interface{}, error)
+	GetWithTTL(key string) (interface{}, time.Duration, error)
+	Set(key string, data interface{}) error
+	SetTTL(ttl time.Duration) error
+	SetWithTTL(key string, data interface{}, ttl time.Duration) error
+	Remove(key string) error
+	Close() error
+	Purge() error
+}
+
 // Cache is a synchronized map of items that can auto-expire once stale
 type Cache struct {
 	mutex                  sync.Mutex
 	ttl                    time.Duration
 	items                  map[string]*item
-	loaderLock             map[string]*sync.Cond
+	loaderLock             *singleflight.Group
 	expireCallback         ExpireCallback
 	expireReasonCallback   ExpireReasonCallback
 	checkExpireCallback    CheckExpireCallback
@@ -266,19 +280,46 @@ func (cache *Cache) SetWithTTL(key string, data interface{}, ttl time.Duration)
 // Get is a thread-safe way to lookup items
 // Every lookup, also touches the item, hence extending it's life
 func (cache *Cache) Get(key string) (interface{}, error) {
+	return cache.GetByLoader(key, nil)
+}
+
+// GetWithTTL has exactly the same behaviour as Get but also returns
+// the remaining TTL for an specific item at the moment it its retrieved
+func (cache *Cache) GetWithTTL(key string) (interface{}, time.Duration, error) {
+	return cache.GetByLoaderWithTtl(key, nil)
+}
+
+// GetByLoader can take a per key loader function (ie. to propagate context)
+func (cache *Cache) GetByLoader(key string, customLoaderFunction LoaderFunction) (interface{}, error) {
+	dataToReturn, _, err := cache.GetByLoaderWithTtl(key, customLoaderFunction)
+
+	return dataToReturn, err
+}
+
+// GetByLoaderWithTtl can take a per key loader function (ie. to propagate context)
+func (cache *Cache) GetByLoaderWithTtl(key string, customLoaderFunction LoaderFunction) (interface{}, time.Duration, error) {
 	cache.mutex.Lock()
 	if cache.isShutDown {
 		cache.mutex.Unlock()
-		return nil, ErrClosed
+		return nil, 0, ErrClosed
 	}
 
 	cache.metrics.Hits++
 	item, exists, triggerExpirationNotification := cache.getItem(key)
 
 	var dataToReturn interface{}
+	ttlToReturn := time.Duration(0)
 	if exists {
 		cache.metrics.Retrievals++
 		dataToReturn = item.data
+		if !cache.skipTTLExtension {
+			ttlToReturn = item.ttl
+		} else {
+			ttlToReturn = time.Until(item.expireAt)
+		}
+		if ttlToReturn < 0 {
+			ttlToReturn = 0
+		}
 	}
 
 	var err error
@@ -286,58 +327,53 @@ func (cache *Cache) Get(key string) (interface{}, error) {
 		cache.metrics.Misses++
 		err = ErrNotFound
 	}
-	if cache.loaderFunction == nil || exists {
+
+	loaderFunction := cache.loaderFunction
+	if customLoaderFunction != nil {
+		loaderFunction = customLoaderFunction
+	}
+
+	if loaderFunction == nil || exists {
 		cache.mutex.Unlock()
 	}
 
-	if cache.loaderFunction != nil && !exists {
-		if lock, ok := cache.loaderLock[key]; ok {
-			// if a lock is present then a fetch is in progress and we wait.
-			cache.mutex.Unlock()
-			lock.L.Lock()
-			lock.Wait()
-			lock.L.Unlock()
-			cache.mutex.Lock()
-			item, exists, triggerExpirationNotification = cache.getItem(key)
-			if exists {
-				dataToReturn = item.data
-				err = nil
-			}
-			cache.mutex.Unlock()
-		} else {
-			// if no lock is present we are the leader and should set the lock and fetch.
-			m := sync.NewCond(&sync.Mutex{})
-			cache.loaderLock[key] = m
-			cache.mutex.Unlock()
-			// cache is not blocked during IO
-			dataToReturn, err = cache.invokeLoader(key)
-			cache.mutex.Lock()
-			m.Broadcast()
-			// cleanup so that we don't block consecutive access.
-			delete(cache.loaderLock, key)
-			cache.mutex.Unlock()
+	if loaderFunction != nil && !exists {
+		type loaderResult struct {
+			data interface{}
+			ttl  time.Duration
 		}
-
+		ch := cache.loaderLock.DoChan(key, func() (interface{}, error) {
+			// cache is not blocked during io
+			invokeData, ttl, err := cache.invokeLoader(key, loaderFunction)
+			lr := &loaderResult{
+				data: invokeData,
+				ttl:  ttl,
+			}
+			return lr, err
+		})
+		cache.mutex.Unlock()
+		res := <-ch
+		dataToReturn = res.Val.(*loaderResult).data
+		ttlToReturn = res.Val.(*loaderResult).ttl
+		err = res.Err
 	}
 
 	if triggerExpirationNotification {
 		cache.expirationNotification <- true
 	}
 
-	return dataToReturn, err
+	return dataToReturn, ttlToReturn, err
 }
 
-func (cache *Cache) invokeLoader(key string) (dataToReturn interface{}, err error) {
-	var ttl time.Duration
-
-	dataToReturn, ttl, err = cache.loaderFunction(key)
+func (cache *Cache) invokeLoader(key string, loaderFunction LoaderFunction) (dataToReturn interface{}, ttl time.Duration, err error) {
+	dataToReturn, ttl, err = loaderFunction(key)
 	if err == nil {
 		err = cache.SetWithTTL(key, dataToReturn, ttl)
 		if err != nil {
 			dataToReturn = nil
 		}
 	}
-	return dataToReturn, err
+	return dataToReturn, ttl, err
 }
 
 // Remove removes an item from the cache if it exists, triggers expiration callback when set. Can return ErrNotFound if the entry was not present.
@@ -369,6 +405,23 @@ func (cache *Cache) Count() int {
 	return length
 }
 
+// GetKeys returns all keys of items in the cache. Returns nil when the cache has been closed.
+func (cache *Cache) GetKeys() []string {
+	cache.mutex.Lock()
+	defer cache.mutex.Unlock()
+
+	if cache.isShutDown {
+		return nil
+	}
+	keys := make([]string, len(cache.items))
+	i := 0
+	for k := range cache.items {
+		keys[i] = k
+		i++
+	}
+	return keys
+}
+
 // SetTTL sets the global TTL value for items in the cache, which can be overridden at the item level.
 func (cache *Cache) SetTTL(ttl time.Duration) error {
 	cache.mutex.Lock()
@@ -385,22 +438,30 @@ func (cache *Cache) SetTTL(ttl time.Duration) error {
 
 // SetExpirationCallback sets a callback that will be called when an item expires
 func (cache *Cache) SetExpirationCallback(callback ExpireCallback) {
+	cache.mutex.Lock()
+	defer cache.mutex.Unlock()
 	cache.expireCallback = callback
 }
 
 // SetExpirationReasonCallback sets a callback that will be called when an item expires, includes reason of expiry
 func (cache *Cache) SetExpirationReasonCallback(callback ExpireReasonCallback) {
+	cache.mutex.Lock()
+	defer cache.mutex.Unlock()
 	cache.expireReasonCallback = callback
 }
 
 // SetCheckExpirationCallback sets a callback that will be called when an item is about to expire
 // in order to allow external code to decide whether the item expires or remains for another TTL cycle
 func (cache *Cache) SetCheckExpirationCallback(callback CheckExpireCallback) {
+	cache.mutex.Lock()
+	defer cache.mutex.Unlock()
 	cache.checkExpireCallback = callback
 }
 
 // SetNewItemCallback sets a callback that will be called when a new item is added to the cache
 func (cache *Cache) SetNewItemCallback(callback ExpireCallback) {
+	cache.mutex.Lock()
+	defer cache.mutex.Unlock()
 	cache.newItemCallback = callback
 }
 
@@ -408,12 +469,16 @@ func (cache *Cache) SetNewItemCallback(callback ExpireCallback) {
 // no longer extend TTL of items when they are retrieved using Get, or when their expiration condition is evaluated
 // using SetCheckExpirationCallback.
 func (cache *Cache) SkipTTLExtensionOnHit(value bool) {
+	cache.mutex.Lock()
+	defer cache.mutex.Unlock()
 	cache.skipTTLExtension = value
 }
 
 // SetLoaderFunction allows you to set a function to retrieve cache misses. The signature matches that of the Get function.
 // Additional Get calls on the same key block while fetching is in progress (groupcache style).
 func (cache *Cache) SetLoaderFunction(loader LoaderFunction) {
+	cache.mutex.Lock()
+	defer cache.mutex.Unlock()
 	cache.loaderFunction = loader
 }
 
@@ -434,6 +499,8 @@ func (cache *Cache) Purge() error {
 // If a new item is getting cached, the closes item to being timed out will be replaced
 // Set to 0 to turn off
 func (cache *Cache) SetCacheSizeLimit(limit int) {
+	cache.mutex.Lock()
+	defer cache.mutex.Unlock()
 	cache.sizeLimit = limit
 }
 
@@ -444,7 +511,7 @@ func NewCache() *Cache {
 
 	cache := &Cache{
 		items:                  make(map[string]*item),
-		loaderLock:             make(map[string]*sync.Cond),
+		loaderLock:             &singleflight.Group{},
 		priorityQueue:          newPriorityQueue(),
 		expirationNotification: make(chan bool),
 		expirationTime:         time.Now(),
@@ -465,6 +532,18 @@ func (cache *Cache) GetMetrics() Metrics {
 	return cache.metrics
 }
 
+// Touch resets the TTL of the key when it exists, returns ErrNotFound if the key is not present.
+func (cache *Cache) Touch(key string) error {
+	cache.mutex.Lock()
+	defer cache.mutex.Unlock()
+	item, exists := cache.items[key]
+	if !exists {
+		return ErrNotFound
+	}
+	item.touch()
+	return nil
+}
+
 func min(duration time.Duration, second time.Duration) time.Duration {
 	if duration < second {
 		return duration
