@@ -20,6 +20,134 @@ func TestMain(m *testing.M) {
 	goleak.VerifyTestMain(m)
 }
 
+// The SimpleCache interface enables quick-start.
+func TestCache_SimpleCache(t *testing.T) {
+	t.Parallel()
+	var cache SimpleCache = NewCache()
+
+	cache.SetTTL(time.Second)
+	cache.Set("k", "v")
+	cache.Get("k")
+	cache.Purge()
+	cache.Close()
+
+}
+
+// Issue 45 : This test was used to test different code paths for best performance.
+func TestCache_GetByLoaderRace(t *testing.T) {
+	t.Skip()
+	t.Parallel()
+	cache := NewCache()
+	cache.SetTTL(time.Microsecond)
+	defer cache.Close()
+
+	loaderInvocations := uint64(0)
+	inFlight := uint64(0)
+
+	globalLoader := func(key string) (data interface{}, ttl time.Duration, err error) {
+		atomic.AddUint64(&inFlight, 1)
+		atomic.AddUint64(&loaderInvocations, 1)
+		time.Sleep(time.Microsecond)
+		assert.Equal(t, uint64(1), inFlight)
+		defer atomic.AddUint64(&inFlight, ^uint64(0))
+		return "global", 0, nil
+
+	}
+	cache.SetLoaderFunction(globalLoader)
+
+	for i := 0; i < 1000; i++ {
+		wg := sync.WaitGroup{}
+		for i := 0; i < 1000; i++ {
+			wg.Add(1)
+			go func() {
+				key, _ := cache.Get("test")
+				assert.Equal(t, "global", key)
+				wg.Done()
+
+			}()
+		}
+		wg.Wait()
+		t.Logf("Invocations: %d\n", loaderInvocations)
+		loaderInvocations = 0
+	}
+
+}
+
+// Issue / PR #39: add customer loader function for each Get() #
+// some middleware prefers to define specific context's etc per Get.
+// This is faciliated by supplying a loder function with Get's.
+func TestCache_GetByLoader(t *testing.T) {
+	t.Parallel()
+	cache := NewCache()
+	defer cache.Close()
+
+	globalLoader := func(key string) (data interface{}, ttl time.Duration, err error) {
+		return "global", 0, nil
+	}
+	cache.SetLoaderFunction(globalLoader)
+
+	localLoader := func(key string) (data interface{}, ttl time.Duration, err error) {
+		return "local", 0, nil
+	}
+
+	key, _ := cache.Get("test")
+	assert.Equal(t, "global", key)
+
+	cache.Remove("test")
+
+	localKey, _ := cache.GetByLoader("test", localLoader)
+	assert.Equal(t, "local", localKey)
+
+	cache.Remove("test")
+
+	globalKey, _ := cache.GetByLoader("test", globalLoader)
+	assert.Equal(t, "global", globalKey)
+
+	cache.Remove("test")
+
+	defaultKey, _ := cache.GetByLoader("test", nil)
+	assert.Equal(t, "global", defaultKey)
+
+	cache.Remove("test")
+}
+
+func TestCache_GetByLoaderWithTtl(t *testing.T) {
+	t.Parallel()
+	cache := NewCache()
+	defer cache.Close()
+
+	globalTtl := time.Duration(time.Minute)
+	globalLoader := func(key string) (data interface{}, ttl time.Duration, err error) {
+		return "global", globalTtl, nil
+	}
+	cache.SetLoaderFunction(globalLoader)
+
+	localTtl := time.Duration(time.Hour)
+	localLoader := func(key string) (data interface{}, ttl time.Duration, err error) {
+		return "local", localTtl, nil
+	}
+
+	key, ttl, _ := cache.GetWithTTL("test")
+	assert.Equal(t, "global", key)
+	assert.Equal(t, ttl, globalTtl)
+	cache.Remove("test")
+
+	localKey, ttl2, _ := cache.GetByLoaderWithTtl("test", localLoader)
+	assert.Equal(t, "local", localKey)
+	assert.Equal(t, ttl2, localTtl)
+	cache.Remove("test")
+
+	globalKey, ttl3, _ := cache.GetByLoaderWithTtl("test", globalLoader)
+	assert.Equal(t, "global", globalKey)
+	assert.Equal(t, ttl3, globalTtl)
+	cache.Remove("test")
+
+	defaultKey, ttl4, _ := cache.GetByLoaderWithTtl("test", nil)
+	assert.Equal(t, "global", defaultKey)
+	assert.Equal(t, ttl4, globalTtl)
+	cache.Remove("test")
+}
+
 // Issue #38: Feature request: ability to know why an expiry has occurred
 func TestCache_textExpirationReasons(t *testing.T) {
 	t.Parallel()
@@ -57,6 +185,50 @@ func TestCache_textExpirationReasons(t *testing.T) {
 
 }
 
+func TestCache_TestTouch(t *testing.T) {
+	t.Parallel()
+	cache := NewCache()
+	defer cache.Close()
+
+	lock := sync.Mutex{}
+
+	lock.Lock()
+	expired := false
+	lock.Unlock()
+
+	cache.SkipTTLExtensionOnHit(true)
+	cache.SetExpirationCallback(func(key string, value interface{}) {
+		lock.Lock()
+		defer lock.Unlock()
+		expired = true
+	})
+
+	cache.SetWithTTL("key", "data", time.Millisecond*900)
+	<-time.After(time.Millisecond * 500)
+
+	// no Touch
+	//	cache.Touch("key")
+
+	<-time.After(time.Millisecond * 500)
+	lock.Lock()
+	assert.Equal(t, true, expired)
+	lock.Unlock()
+	cache.Remove("key")
+
+	lock.Lock()
+	expired = false
+	lock.Unlock()
+
+	cache.SetWithTTL("key", "data", time.Millisecond*900)
+	<-time.After(time.Millisecond * 500)
+	cache.Touch("key")
+
+	<-time.After(time.Millisecond * 500)
+	lock.Lock()
+	assert.Equal(t, false, expired)
+	lock.Unlock()
+}
+
 // Issue #37: Cache metrics
 func TestCache_TestMetrics(t *testing.T) {
 	t.Parallel()
@@ -183,6 +355,37 @@ func TestCache_TestLoaderFunctionDuringClose(t *testing.T) {
 
 }
 
+// Cache sometimes returns key not found under parallel access with a loader function
+func TestCache_TestLoaderFunctionParallelKeyAccess(t *testing.T) {
+	t.Parallel()
+	cache := NewCache()
+
+	cache.SetLoaderFunction(func(key string) (data interface{}, ttl time.Duration, err error) {
+		time.Sleep(time.Millisecond * 300)
+		return "1", 1 * time.Nanosecond, nil
+	})
+
+	wg := sync.WaitGroup{}
+	errCount := uint64(0)
+	for i := 0; i < 200; i++ {
+		wg.Add(1)
+		go func() {
+			defer wg.Done()
+			value, found := cache.Get("foo")
+			if value != "1" || found != nil { // Use an atomic to avoid spamming logs
+				atomic.AddUint64(&errCount, 1)
+			}
+		}()
+
+	}
+
+	wg.Wait()
+
+	assert.Equalf(t, uint64(0), errCount, "expected 0 errs, got %d", errCount)
+
+	cache.Close()
+}
+
 // Issue #28: call expirationCallback automatically on cache.Close()
 func TestCache_ExpirationOnClose(t *testing.T) {
 	t.Parallel()
@@ -611,6 +814,88 @@ func TestCacheGet(t *testing.T) {
 	assert.Equal(t, "world", (data.(string)), "Expected data content to be 'world'")
 }
 
+func TestCacheGetKeys(t *testing.T) {
+	t.Parallel()
+
+	cache := NewCache()
+	defer cache.Close()
+
+	keys := cache.GetKeys()
+	assert.Empty(t, keys, "Expected keys to be empty")
+
+	cache.Set("hello", "world")
+	keys = cache.GetKeys()
+	assert.NotEmpty(t, keys, "Expected keys to be not empty")
+	assert.Equal(t, []string{"hello"}, keys, "Expected keys contains 'hello'")
+}
+
+func TestCacheGetWithTTL(t *testing.T) {
+	t.Parallel()
+
+	cache := NewCache()
+	defer cache.Close()
+
+	data, ttl, exists := cache.GetWithTTL("hello")
+	assert.Equal(t, exists, ErrNotFound, "Expected empty cache to return no data")
+	assert.Nil(t, data, "Expected data to be empty")
+	assert.Equal(t, int(ttl), 0, "Expected item TTL to be 0")
+
+	cache.Set("hello", "world")
+	data, ttl, exists = cache.GetWithTTL("hello")
+	assert.NotNil(t, data, "Expected data to be not nil")
+	assert.Equal(t, nil, exists, "Expected data to exist")
+	assert.Equal(t, "world", (data.(string)), "Expected data content to be 'world'")
+	assert.Equal(t, int(ttl), 0, "Expected item TTL to be 0")
+
+	orgttl := time.Duration(500 * time.Millisecond)
+	cache.SetWithTTL("hello", "world", orgttl)
+	time.Sleep(10 * time.Millisecond)
+	data, ttl, exists = cache.GetWithTTL("hello")
+	assert.NotNil(t, data, "Expected data to be not nil")
+	assert.Equal(t, nil, exists, "Expected data to exist")
+	assert.Equal(t, "world", (data.(string)), "Expected data content to be 'world'")
+	assert.Equal(t, ttl, orgttl, "Expected item TTL to be original TTL")
+
+	cache.SkipTTLExtensionOnHit(true)
+	cache.SetWithTTL("hello", "world", orgttl)
+	time.Sleep(10 * time.Millisecond)
+	data, ttl, exists = cache.GetWithTTL("hello")
+	assert.NotNil(t, data, "Expected data to be not nil")
+	assert.Equal(t, nil, exists, "Expected data to exist")
+	assert.Equal(t, "world", (data.(string)), "Expected data content to be 'world'")
+	assert.Less(t, ttl, orgttl, "Expected item TTL to be less than the original TTL")
+	assert.NotEqual(t, int(ttl), 0, "Expected item TTL to be not 0")
+}
+
+func TestCache_TestGetWithTTLAndLoaderFunction(t *testing.T) {
+	t.Parallel()
+	cache := NewCache()
+
+	cache.SetLoaderFunction(func(key string) (data interface{}, ttl time.Duration, err error) {
+		return nil, 0, ErrNotFound
+	})
+
+	_, ttl, err := cache.GetWithTTL("1")
+	assert.Equal(t, ErrNotFound, err, "Expected error to be ErrNotFound")
+	assert.Equal(t, int(ttl), 0, "Expected item TTL to be 0")
+
+	orgttl := time.Duration(1 * time.Second)
+	cache.SetLoaderFunction(func(key string) (data interface{}, ttl time.Duration, err error) {
+		return "1", orgttl, nil
+	})
+
+	value, ttl, found := cache.GetWithTTL("1")
+	assert.Equal(t, nil, found)
+	assert.Equal(t, "1", value)
+	assert.Equal(t, ttl, orgttl, "Expected item TTL to be the same as the original TTL")
+	cache.Close()
+
+	value, ttl, found = cache.GetWithTTL("1")
+	assert.Equal(t, ErrClosed, found)
+	assert.Equal(t, nil, value)
+	assert.Equal(t, int(ttl), 0, "Expected returned ttl for an ErrClosed err to be 0")
+}
+
 func TestCacheExpirationCallbackFunction(t *testing.T) {
 	t.Parallel()
 
